
/* HardDisk.cpp - HardDisk API Abstraction Class
 *
 * Copyright (c) 2008 Topalis AG
 * Author: Christian Kellner <christian.kellner@topalis.com>
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor,
 * Boston, MA  02110-1301, USA.
 */

#include <VBox/VBoxHDD.h>

#include <cstdio>
#include <map>
#include <iostream>

#include "IOError.h"
#include "VBoxFailure.h"
#include "UUID.h"

#include <iprt/stdint.h>
#include <VBox/err.h> /* VBOX_FAILURE */
#include "HardDisk.h"

#define printf RTPrintf

void
HardDisk::OpenImages (std::list<std::string> &pathList)
{
    std::list<std::string> sorted = SortFamily (pathList);

    std::list<std::string>::iterator iter;
    for (iter = sorted.begin (); iter != sorted.end (); iter++) {
        OpenImage (*iter, DefaultOpenFlags ());
    }
}

Buffer *
HardDisk::Read (uint64_t offset, size_t len)
{
    Buffer *buffer = new Buffer (len);
    void *data = buffer->Reserve (len);

    Read (offset, data, len);

    return buffer;
}

void
HardDisk::Write (uint64_t offset, Buffer* const buffer)
{
    size_t len;

    const void *data = buffer->GetData (&len);
    Write (offset, data, len);
}

std::list<std::string>
HardDisk::SortFamily (std::list<std::string> &pathList)
{
    std::map<UUID, std::string> u2p;
    std::map<UUID, UUID> p2c;

    std::list<std::string>::iterator iter;
    for (iter = pathList.begin ();
         iter != pathList.end ();
         iter++) {

        std::string &curPath = *iter;
        RTUUID myid, paid;
        int vc = VDIGetImageUUIDs (curPath.c_str (),
                                   &myid, NULL, &paid,
                                   NULL);

        if (VBOX_FAILURE (vc)) {
            throw VBoxFailure (vc);
        }

        //p2c::iterator pos;
        UUID parent = paid;
        UUID child = myid;

        u2p.insert (std::make_pair (parent, curPath));
        p2c.insert (std::make_pair (parent, child));
    }

    std::list<std::string> sorted;

    /* Find the root */
    UUID child = UUID::Null;
    std::map<UUID, UUID>::iterator pos;

    while ((pos = p2c.find (child)) != p2c.end ()) {
        const UUID &parent = pos->first;

        std::string path = u2p[parent];
        sorted.push_back (path);
        child = pos->second;
    }

    if (sorted.size () != pathList.size ()) {
        std::cerr << "Missing Link!: " << child << "\n";
        throw 1;
    }

    return sorted;
}
