
/* HardDiskNew.cpp - New VDI API
 *
 * Copyright (c) 2008 Topalis AG
 * Author: Christian Kellner <christian.kellner@topalis.com>
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor,
 * Boston, MA  02110-1301, USA.
 */

#include "Log.h"

#include <iprt/runtime.h>
#include <iprt/stream.h>
#include <iprt/string.h>
#include <iprt/asm.h>
#include <iprt/uuid.h>
#include <iprt/thread.h>
#include <iprt/path.h>
#include <iprt/param.h>
#include <iprt/dir.h>
#include <iprt/file.h>
#include <iprt/env.h>
#include <iprt/cidr.h>
#include <VBox/err.h>
#include <VBox/version.h>

#include <cstdio>


#include "IOError.h"
#include "HardDiskNew.h"

#define printf RTPrintf

HardDiskNew::HardDiskNew ()
    : diskHandle (NULL), nImages (0), format (NULL)
{

}
HardDiskNew::~HardDiskNew ()
{
    if (diskHandle) {
        Close ();
    }
}

static
void handleVDError (void *pvUser, int rc, RT_SRC_POS_DECL, const char *format, va_list va)
{
    RTPrintf ("ERROR: ");
    RTPrintfV (format, va);
    RTPrintf ("\n");
    RTPrintf ("Error code %Vrc at %s(%u) in function %s\n", rc, RT_SRC_POS_ARGS);
}

int
HardDiskNew::Initilize ()
{
    int vrc;

    vdInterfaceErrorCallbacks.cbSize       = sizeof (VDINTERFACEERROR);
    vdInterfaceErrorCallbacks.enmInterface = VDINTERFACETYPE_ERROR;
    vdInterfaceErrorCallbacks.pfnError     = handleVDError;

    vrc = VDInterfaceAdd (&vdInterfaceError, "HardDiskNew_IError", VDINTERFACETYPE_ERROR,
                          &vdInterfaceErrorCallbacks, this, &vdIfs);

    vrc = VDCreate (&vdInterfaceError, &diskHandle);

    if (VBOX_FAILURE (vrc)) {
        diskHandle = NULL; /* just to be sure */
    }

    return vrc;
}

void
HardDiskNew::OpenImage (std::string path, unsigned int flags)
{
    if (diskHandle == NULL) {
        int vrc = Initilize ();

        if (VBOX_FAILURE (vrc)) {
            throw VBoxFailure (vrc);
        }
    }

    int vrc = 0;

    if (nImages == 0) {
        vrc = VDGetFormat (path.c_str (), &format);

        if (VBOX_FAILURE (vrc)) {
            LogWarning (("Error while gettin the disk format: %Vrc\n", vrc));
            throw VBoxFailure (vrc);
        }
    }

    vrc = VDOpen (diskHandle, format, path.c_str (), flags, NULL);

    if (VBOX_FAILURE (vrc)) {
        throw VBoxFailure (vrc);
    }

    nImages++;
}

void
HardDiskNew::Read (uint64_t offset, void *data, size_t len)
{
    int vrc = VDRead (diskHandle, offset, data, len);

    if (VBOX_FAILURE (vrc))
        throw VBoxFailure (vrc);
}



void
HardDiskNew::Write (uint64_t offset, const void *data, size_t len)
{
    int vrc = VDWrite (diskHandle, offset, data, len);

    if (VBOX_FAILURE (vrc))
        throw VBoxFailure (vrc);
}

uint64_t
HardDiskNew::GetSize ()
{
    return VDGetSize (diskHandle, VD_LAST_IMAGE);
}

void
HardDiskNew::Close ()
{
    if (diskHandle) {
       VDDestroy (diskHandle);
       diskHandle = NULL;
    }
}

bool
HardDiskNew::IsReadOnly ()
{
    return VDIsReadOnly (diskHandle);
}

const char *
HardDiskNew::GetFormat ()
{
    return format;
}
