
/* Nbd.cpp - Network Block Device Abstraction Class
 *
 * Copyright (c) 2008 Topalis AG
 * Author: Christian Kellner <christian.kellner@topalis.com>
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor,
 * Boston, MA  02110-1301, USA.
 */

#include "Log.h"

#include <VBox/err.h>

#include <sys/types.h>
#include <sys/socket.h>
#include <sys/un.h>
#include <sys/vfs.h>
#include <sys/time.h>
#include <sys/ioctl.h>
#include <sys/stat.h>
#include <sys/wait.h>

#include <arpa/inet.h>

#include <cerrno>

#include <linux/fs.h>
#include <byteswap.h>

#include <linux/nbd.h>


#include "HardDisk.h" // For VBoxFailure only

#ifdef WORDS_BIGENDIAN //FIXME
uint64_t
static ntohll (uint64_t a) {
	return a;
}
#else
uint64_t
static ntohll (uint64_t val) {
    return bswap_64 (val);
}
#endif
#define htonll ntohll

#include "Nbd.h"

void
Nbd::Open (int flags)
{
    int res = open (devicePath.c_str (), flags);

    if (res < 0) {
        throw IOError ("Error opening NBD", errno);
    }

    fd = res;
}

void
Nbd::ClearQueue (bool ignoreErrors)
{
    int res;

    res = ioctl (fd, NBD_CLEAR_QUE);

    if (ignoreErrors == false && res < 0)
        throw IOError ("Nbd::ClearQueue", errno);
}

void
Nbd::Disconnect ()
{
    int res;

    res = ioctl (fd, NBD_DISCONNECT);

    if (res < 0)
        throw IOError ("Nbd::Disconnect", errno);
}

void
Nbd::ClearSocket (bool ignoreErrors)
{
    int res;

    res = ioctl (fd, NBD_CLEAR_SOCK);

    if (ignoreErrors == false && res < 0)
        throw IOError ("Nbd::ClearSocket", errno);
}

void
Nbd::SetBlockSize (unsigned long blocksize)
{
    int res;

    res = ioctl (fd, NBD_SET_BLKSIZE, blocksize);

    if (res < 0)
        throw IOError ("Nbd::SetBlockSize", errno);

}

void
Nbd::SetSizeBlocks (unsigned long blocks)
{
    int res;

    res = ioctl (fd, NBD_SET_SIZE_BLOCKS, blocks);

    if (res < 0)
        throw IOError ("Nbd::SetSizeBlocks", errno);

}

void Nbd::SetReadOnly (bool ro)
{
    int res;
    unsigned long value = ro ? 1 : 0;

    res = ioctl (fd, BLKROSET, &value);

    if (res < 0)
        throw IOError ("Nbd::SetReadOnly", errno);
}

void Nbd::SetTimeOut (unsigned long timeout)
{
    int res = ioctl (fd, NBD_SET_TIMEOUT, timeout);

    if (res < 0)
        throw IOError ("Nbd::SetTimeOut", errno);
}

void
Nbd::SetSocket (int socket)
{
    int res;

    res = ioctl (fd, NBD_SET_SOCK, socket);

    if (res < 0)
        throw IOError ("Nbd::SetSocket", errno);

    sock = socket;
}

int
Nbd::DoIt ()
{
    return ioctl (fd, NBD_DO_IT);
}

static int
nbd_doit_thread (RTTHREAD ThreadSelf, void *userData)
{
    LogFlowFuncEnter ();

    Nbd *nbd = reinterpret_cast<Nbd *> (userData);

restart:
    int res = nbd->DoIt ();

    LogRel (("Nbd->DoIt: %d (%s) \n", res, strerror (errno)));

    if (res < 0) {

        if (errno == EINTR) {
            goto restart;
        } else if (errno == EPIPE) {

            /* Most likely normal shutdown, don't complain */
            LogRel (("Nbd->Doit (Thread) returned EPIPE\n"));
            nbd->ClearSocket ();

            res = nbd->Close ();

        } else {
            LogWarning (("DO_IT RETUNRED: %s\n",
                         strerror (errno)));
            nbd->ClearSocket ();
            nbd->Close (); /* ignore the clsoe value */
        }

    }  else {
        LogWarningFunc (("Not sure we should ever land here\n"));
        nbd->ClearSocket ();
        res = nbd->Close ();
    }

    LogFlowFunc (("LEAVE (res: %d)\n", res));
    return res;
}

void Nbd::DoItBegin ()
{
    LogFlowThisFuncEnter ();

    int rc = RTThreadCreate (&doItThread, nbd_doit_thread, this,
                             0, RTTHREADTYPE_MAIN_WORKER,
                             RTTHREADFLAGS_WAITABLE, "NbdDoIt");

    if (VBOX_FAILURE (rc)) {
        throw VBoxFailure (rc);
    }

    LogFlowThisFuncLeave ();
}

int
Nbd::DoItEnd ()
{
    int res = -1;
    LogFlowThisFuncEnter ();

    RTThreadWait (doItThread, RT_INDEFINITE_WAIT, &res);

    Log (("RTThreadWait returned: %d\n", res));
    LogFlowThisFuncEnter ();
    return res;

}


int
Nbd::PokeThread (RTTHREAD ThreadSelf, void *userData)
{
    Nbd *nbd = reinterpret_cast<Nbd *> (userData);

    int fd = open (nbd->devicePath.c_str (), O_RDONLY);

    if (fd < 0) {
        return 1;
    }

    int res = close (fd);
    return res;
}

int
Nbd::Poke ()
{
    int rc = RTThreadCreate (&doItThread, Nbd::PokeThread, this,
                             0, RTTHREADTYPE_MAIN_WORKER,
                             0, "Nbd::Poke");
    return rc;
}

void
Nbd::PokeSync (int timeout)
{
    char buf[32];

    int fd = open (devicePath.c_str (), O_RDONLY);

    if (fd < 0) {
        throw IOError ("Nbd::PokeSync::Open", errno);
    }

    Socket sock = fd;
    sock.SetTimeOut (timeout);
    sock.Read (&buf, sizeof (buf));
    sock.Close ();
}

int
Nbd::Close ()
{
    int res = 1;
    if (fd > 0) {
        res = close (fd);
        fd = -1;
    }

    if (sock) {
        close (sock);
    }

    return res;
}

Nbd::~Nbd ()
{
    if (fd > 0)
        close (fd);
}


/* ***************************************************** */
Nbd::Request::Request (Socket &sock, char handle[8])
    : sock (sock)
{
    memcpy (this->handle, handle, sizeof (this->handle));
}

Nbd::Request
Nbd::Request::RecieveNext (Socket &sock)
{
    static uint32_t magic = htonl (NBD_REQUEST_MAGIC);
    struct nbd_request nbd_req;

    sock.Read (&nbd_req, sizeof (nbd_req));

    if (nbd_req.magic != magic) {
        throw IOError ("Proto Error (Bad Request magic)", EIO);
        /* Should be Protocol Error (FIXME) */
    }

    Request request = Request (sock, nbd_req.handle);
    switch (ntohl (nbd_req.type)) {
        case NBD_CMD_DISC:
            request.type = CmdDisconnect;
            break;
        case NBD_CMD_WRITE:
            request.type = CmdWrite;
            break;
        case NBD_CMD_READ:
            request.type = CmdRead;
            break;
        default:
            throw IOError ("Proto Error (Unknown Cmd)", EIO);
    }

    request.lenght = ntohl (nbd_req.len);
    request.offset = ntohll (nbd_req.from);

    return request;
}

Buffer *
Nbd::Request::GetPayload (void)
{
    Buffer *buffer = sock.Read (lenght);
    return buffer;
}

Nbd::Reply *
Nbd::Request::CreateReply ()
{
    Reply *reply = new Reply (sock, handle);

    return reply;
}


/* ********* */
Nbd::Reply::Reply (Socket &sock, char handle[8])
    : error (0), sock (sock), payload (NULL)
{
    memcpy (this->handle, handle, sizeof (this->handle));
}

void
Nbd::Reply::Send ()
{
    static uint32_t magic = htonl (NBD_REPLY_MAGIC);
    struct nbd_reply reply;

    reply.magic = magic;
    memcpy (reply.handle, handle, sizeof (reply.handle));
    reply.error = htonl (error);

    ssize_t res = sock.Write (&reply, sizeof (reply));

    if (res != sizeof (reply))
        throw IOError ("Nbd::Reply::Send", res);

    if (payload != NULL)
        sock.Write (payload);
}

void
Nbd::Reply::SetPayload (Buffer *data)
{
    this->payload = data;
}
