
/* Socket.cpp - Socket Abstraction Class
 *
 * Copyright (c) 2008 Topalis AG
 * Author: Christian Kellner <christian.kellner@topalis.com>
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor,
 * Boston, MA  02110-1301, USA.
 */

#include <unistd.h>
#include <errno.h>


#include "Socket.h"
#include "IOError.h"

#include <fcntl.h>

int
Socket::Wait (bool read)
{
    fd_set fds;
    int max_fds;

    struct timeval tout;
    struct timeval *toutp = NULL;
    FD_ZERO (&fds);
    FD_SET (sock, &fds);

    if (timeout > 0) {
        tout.tv_sec = timeout;
        tout.tv_usec = 0;
        toutp = &tout;
    }

    max_fds = sock;
    int res;

    LogRel (("Wainting ...\n"));

    if (read)
        res = select (max_fds + 1, &fds, NULL, NULL, toutp);
    else
        res = select (max_fds + 1, NULL, &fds, NULL, toutp);

    if (res == -1) {
        throw IOError ("Socket::Wait", errno);
    } else if (res == 0) {
        //XXX Timeout
        throw IOError ("Socket::Wait Timeout", ETIMEDOUT);
    }

    return res;
}

ssize_t
Socket::Read (void *buf, size_t count)
{
    ssize_t len;
    char *p;

    p = (char *) buf;
    len = count;

    while (len > 0) {
        ssize_t res;
        res = read (sock, p, len);

        if (res == -1) {
            if (errno == EINTR)
                continue;
            else if (errno == EAGAIN) {
                Wait (true);
                continue;
            }

            throw IOError ("Socket::Read", errno);
        }

        len -= res;
        p += res;

        if (res == 0 && len) {
            throw IOError ("Socket::Read (all)", EOF);
        }

    }

    return count;
}

ssize_t
Socket::Write (const void *buf, size_t count)
{
    ssize_t res;
    ssize_t len;
    char *p;

    len = count;
    p = (char *) buf;

    while (len > 0) {

        res = write (sock, p, len);

        if (res == -1) {
            if (errno == EINTR) {
                continue;
            } else if (errno == EAGAIN) {
                Wait (false);
                continue;
            }

            throw IOError ("Socket::Write", errno);
        }

        len -= res;
        p += res;

        if (res == 0)
            return count - len;
    }

    return count;
}

Buffer *
Socket::Read (size_t count)
{
    Buffer *buffer = new Buffer (count);
    void *data = buffer->Reserve (count);

    ssize_t n = Read (data, count);

    if ((size_t) n < count) {
        throw IOError ("Socket::Read (Buffer)", EIO);
    }

    return buffer;
}

void
Socket::Write (Buffer* const buffer)
{
    size_t   len;

    const void *data = buffer->GetData (&len);
    ssize_t n = Write (data, len);

    if ((size_t) n < len) {
        throw IOError ("Socket::Write (Buffer)", EIO);
    }
}

int
Socket::Close ()
{
    int res;
    res = close (sock);
    sock = -1;
    return res;
}

void
Socket::SetTimeOut (int secs)
{
    int flags;

    flags = fcntl (sock, F_GETFL, 0);

    if (flags == -1)
        flags = 0;

    fcntl (sock, F_SETFL, flags | O_NONBLOCK);
    timeout = secs;
}

