
/* VBoxMount.cpp - Main File
 *
 * Copyright (c) 2008 Topalis AG
 * Author: Christian Kellner <christian.kellner@topalis.com>
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor,
 * Boston, MA  02110-1301, USA.
 */

#include "Log.h"

#include <iprt/runtime.h>
#include <iprt/process.h>
#include <iprt/env.h>
#include <VBox/err.h>
#include <VBox/version.h>

#include <sys/types.h>
#include <sys/socket.h>
#include <sys/wait.h>
#include <signal.h>

#include <unistd.h>
#include <getopt.h>

#include <cstdarg>
#include <cstdlib>
#include <cmath>
#include <cerrno>
#include <cstring>

#include <string>

#include "Socket.h"
#include "Buffer.h"
#include "HardDiskOld.h"
#include "HardDiskNew.h"
#include "Nbd.h"

/* *********************************************************************** */
static Nbd::Reply *
handle_block_read (Nbd::Request &req, HardDisk *hdd)
{
    Nbd::Reply *reply = req.CreateReply ();

    if (req.lenght + req.offset > hdd->GetSize ()) {
        LogWarning (("Out of bounds request\n"));
        reply->error = EINVAL;
        return reply;
    }

    try {

        Buffer *data = hdd->Read (req.offset, req.lenght);
        reply->SetPayload (data);

    } catch (VBoxFailure &failure) {
        reply->error = EIO;
    }

    return reply;
}

static Nbd::Reply *
handle_block_write (Nbd::Request &req, HardDisk *hdd)
{
    Nbd::Reply *reply = req.CreateReply ();
    Buffer *data = req.GetPayload ();

    try {
        hdd->Write (req.offset, data);
    } catch (VBoxFailure &failure) {
        reply->error = EIO;
    }

    delete data;
    return reply;
}

static bool
processRequest (Nbd::Request &req, HardDisk *hdd)

{
    Nbd::Reply *reply = NULL;

    if (req.type == Nbd::Request::CmdDisconnect) {

        LogRel (("Should Disconnect!\n"));

    } else if (req.type == Nbd::Request::CmdWrite) {

        reply = handle_block_write (req, hdd);

    } else if (req.type == Nbd::Request::CmdRead) {

        reply = handle_block_read (req, hdd);
    }

    if (reply == NULL) {
        return false;
    }

    reply->Send ();
    delete reply;

    return true;
}

/* *********************************************************************** */
static void
show_help ()
{
    printf ("VirtualBox Mount [%s %s] (C) 2008 Toplais AG\n",
            __DATE__, __TIME__);

    printf ("Usage: VBoxMount <nbd device> <DiskImage> [DiskImage...] \n");
    printf ("\nOptions:\n");
    printf ("  -h, --help\t\tShow this\n");
    printf ("  -u, --unmount <nbd device>\tUnmount DiskImage\n");
    printf ("  -w, --write-only\t\tFail if disk was opened readonly\n");
    printf ("  -x, --legacy-api\t\tUse the old VDI API\n");
    printf ("  -l, --logfile <File>\t\tLog to that file\n");
    printf ("  -f, --forground\t\tDon't daemonize, stay in forground\n");
    printf ("\n");
}

static int
do_umount (const char *dev)
{
    Nbd nbd = dev;

    try {

        nbd.Open ();
        nbd.ClearQueue ();
        nbd.Disconnect ();
        nbd.ClearSocket ();

    } catch (IOError &error) {
        LogRel (("Error: %s\n", error.what ()));
        return EXIT_FAILURE;
    }

    return 0;
}
/* *********************************************************************** */

class Daemonizer {
public:
    enum Identity {
        I_PARENT = 0,
        I_CHILD = 1
    };

    Identity Begin   ();
    Identity GetIdentity () { return id; };

    int Finish (int ourStatus);

private:
    int      ipc;
    Identity id;
};


Daemonizer::Identity
Daemonizer::Begin ()
{
    int sp[2];
    int res = socketpair (AF_UNIX, SOCK_STREAM, 0, sp);

    if (res != 0) {
        throw IOError ("Daemonize failed", errno);
    }

    /* */

    pid_t pid = fork ();

    if (pid < 0) {
        throw IOError ("Fork", errno);
    } else if (pid == 0) {
        /* (Intermediate) child */
        close (sp[I_PARENT]);
        ipc = sp[I_CHILD];
        id = I_CHILD;

        int dnfd = open ("/dev/null", O_RDWR);

        if (dnfd != -1) {
            dup2 (dnfd, STDIN_FILENO);
            dup2 (dnfd, STDOUT_FILENO);
            dup2 (dnfd, STDERR_FILENO);
        } else {
            close (STDIN_FILENO);
            close (STDOUT_FILENO);
            close (STDERR_FILENO);
            //XXX make sure we dont log to any of those now
        }

        pid_t gpid = fork ();

        if (gpid != 0) {
            exit (gpid > 0 ? 0 : errno);
        }

        /* GrandChild */
        setsid ();
        setpgid (0, 0);
        res = chdir ("/");
        return I_CHILD;
    }

    /* PARENT (MAIN) */
    /* Reap the intermediate child */
    close (sp[I_CHILD]);
    ipc = sp[I_PARENT];
    id = I_PARENT;

wait_again:
    int status = 0;

    if (waitpid (pid, &status, 0) < 0) {
        if (errno == EINTR) {
            goto wait_again;
        } else if (errno == ECHILD) {
        } else {
        }
    }

    if (status != 0)
        throw IOError ("Daemoize failed (Child)", status);

    return I_PARENT;
}

int
Daemonizer::Finish (int ourStatus)
{
    Socket s = ipc;
    s.Write (&ourStatus, sizeof (ourStatus));
    int status;
    s.Read (&status, sizeof (status));
    s.Close ();
    return status;
}

/* *********************************************************************** */

int
main (int argc, char **argv)
{
    int   opt_index;
    bool  wonly = false;
    bool  lapi = false;
    bool  forground = false;
    char *logfile = NULL;
    const char *nbd_dev = NULL;
    static struct option options[] = {
        {"forground",  no_argument,       0,      'f'},
        {"legacy-api", no_argument,       0,      'x'},
        {"logfile",    required_argument, 0,      'l'},
        {"help",       no_argument,       0,      'h'},
        {"unmount",    required_argument, 0,      'u'},
        {"write-only", no_argument,       0,      'w'},
        {0,            0,                 0,       0 }
    };

    opt_index = 0;
    int res = 0;
    while (1) {

        int c = getopt_long (argc, argv, "fl:u:hwx", options, &opt_index);

        if (c == -1)
            break;

        switch (c) {

            case 0:
                break;

            case 'x':
                lapi = true;
                break;

            case 'l':
                logfile = optarg;
                break;

            case 'f':
                forground = true;
                break;

            case 'u':
                res = do_umount (optarg);
                return res;

            case 'h':
                show_help ();
                return 0;

            case 'w':
                wonly = true;
                break;

            default:
                show_help ();
                return 1;
        }
    }

    if ((argc - optind) < 2) {
        show_help ();
        return 1;
    }

    /* Engines startup ... */
    int rc = RTR3Init ();

    if (VBOX_FAILURE (rc)) {
        RTPrintf ("Could not init runtime\n");
        return 1;
    }

    /* Setup Logging */
    PRTLOGGER Logger;
    static const char * const logGroups[] = VBOX_LOGGROUP_NAMES;
    if (logfile == NULL) {
        RTLogCreate (&Logger, 0, "all",
                     NULL, ELEMENTS(logGroups), logGroups,
                     RTLOGDEST_STDOUT, NULL);

        RTLogRelSetDefaultInstance (Logger);

    } else {
        rc = RTLogCreate (&Logger, 0, "all",
                          NULL, ELEMENTS(logGroups), logGroups,
                          RTLOGDEST_FILE, logfile);

        RTTIMESPEC TimeSpec;
        char szNowUct[64];
        RTTimeSpecToString(RTTimeNow(&TimeSpec), szNowUct, sizeof(szNowUct));

        RTLogRelLogger (Logger, 0, ~0U,
                        "VBoxMount %s (%s %s) release log\n"
                        "Log opened %s\n",
                        VBOX_VERSION_STRING, __DATE__, __TIME__,
                        szNowUct);

        if (VBOX_FAILURE (rc)) {
            RTPrintf ("Warning: Could not create Logger: %Vrc!\n", rc);
        } else {
            RTLogRelSetDefaultInstance (Logger);
        }
    }

    /* Now memoriz the nbd device */
    nbd_dev = argv[optind++];

    /* Setup the HDD */
    HardDisk *hdd;
    if (lapi == false) {
        hdd = new HardDiskNew ();
    } else {
        hdd = new HardDiskOld ();
    }

    try {

        if ((argc - optind) == 1) {
            hdd->OpenImage (argv[optind], hdd->DefaultOpenFlags ());
        } else {
            std::list<std::string> pathList;

            while (argc > optind) {
                pathList.push_back (argv[optind]);
                optind++;
            }

            hdd->OpenImages (pathList);
        }

    } catch (VBoxFailure &failure) {
        LogRel (("Error opening Hdd: %s\n", failure.what ()));
        return 1;
    }

    uint64_t size = hdd->GetSize ();
    const char *format = hdd->GetFormat ();
    bool readOnly = hdd->IsReadOnly ();

    LogRel (("[Hdd Opened: Size: %llu, Format: %s, Read Only: %s]\n",
              size, format, readOnly ? "Yes" : "No"));

    if (wonly && readOnly) {
        LogRel (("Couldn't open HDD read-write.\n"
                 "Failing as requested\n"));
        return -1;
    }

    /* Setup NBD now */
    int sock[2];

    res = socketpair (AF_UNIX, SOCK_STREAM, 0, sock);

    if (res != 0) {
        LogRel (("Socketpair call failed: %s\n",
                 strerror (errno)));
        return 1;
    }

    /* Now create the nbd device */
    Nbd nbd = nbd_dev;

    try {
        unsigned long blocksize = 512;
        unsigned long blocks;

        nbd.Open ();

        nbd.SetBlockSize (blocksize);
        blocks = (unsigned long) (size / blocksize);
        nbd.SetSizeBlocks (blocks);
        nbd.SetReadOnly (readOnly);
        nbd.ClearSocket ();
        //nbd.SetTimeOut (1);
        nbd.SetSocket (sock[1]);

    } catch (IOError &error) {
        LogRel (("Error: %s\n", error.what ()));
        return EXIT_FAILURE;
    }

    /* Submarine mode: on (do here becuase ioctls
     * wont survive forks () due to SIGCHLD
     */
    Daemonizer daemonizer;
    if (forground == false) {
        Daemonizer::Identity id;

        try {

            id = daemonizer.Begin ();

            if (id == Daemonizer::I_PARENT) {
                close (sock[0]);
                int status = daemonizer.Finish (0);
                /* PARENT EXIT */
                fprintf (stderr, "Child Status: %d\n", status);
                return status;
            }

            /* CHILD!! */
            close (sock[1]);

        } catch (VBoxFailure &failure) {
            LogRel (("Error spawning child: %s\n", failure.what ()));
            return 1;
        } catch (IOError &error) {
            LogRel (("Error spawning child: %s\n", error.what ()));
            return 1;
        }
    }

    LogRel (("Calling Do it\n"));
    /* Async Nbd.DoIt () */
    nbd.DoItBegin ();

    /* That should re-read the partition table */
    nbd.Poke ();

    Socket socket = sock[0];
    Buffer buffer;
    bool have_error = false;


    if (forground == false) {
        res = 0;

        /* In case we should daemonize we process the first request
         * and see if we are all good */

        try {
            Nbd::Request request = Nbd::Request::RecieveNext (socket);
            bool res = processRequest (request, hdd);

            if (res == false)
                res = 2;

        } catch (...) {
            res = 1;
        }

        int status = daemonizer.Finish (res);

        if (res != 0 || status != 0) {
            have_error = true;
            goto out;
        }
    }

    try {
        bool cont;

        /* This is our main loop */
        do {

            Nbd::Request request = Nbd::Request::RecieveNext (socket);
            cont = processRequest (request, hdd);

        } while (cont == true);

    } catch (IOError &error) {

        LogRel (("Error: %s\n", error.what ()));
        have_error = true;

    } catch (...) {

        LogRel (("Error: Unkown Exception occured\n"));
        have_error = true;
    }

out:
    int child_status;
    Log (("Closing socket\n"));
    close (sock[0]);

    child_status = nbd.DoItEnd ();

    LogRel (("Have Error: %s (%d)\n",
            have_error ? "yes": "false", child_status));

    RTLogFlush (NULL);

    delete hdd;
    return have_error ? EXIT_FAILURE : EXIT_SUCCESS;
}
